#include "stdafx.h"

double normalizeAngle(double angle)
{
	while (angle < 0)
		angle += 2*PI;
	while (angle > 2*PI)
		angle -= 2*PI;
	return angle;
}

int insideAngles(double phi1, double beta, double phi2)
{
	phi1 = normalizeAngle(phi1);
	phi2 = normalizeAngle(phi2);
	beta = normalizeAngle(beta);
	return phi1 <= phi2 ?  phi1 <= beta && beta <= phi2 : beta >= phi1 || beta <= phi2;
}

ZRK::ZRK(const Vector3D& location, zrc TTX, double importance, double life):ActiveMean(location,importance,life)
{
	//  
	num_channels = TTX.num_channels;
	maxR = TTX.maxR;
	minR = TTX.minR;
	gamma = TTX.gamma / 180.0 * PI;
	shift_fire = TTX.shift_fire;
	rocketV = TTX.rocketV * dt;
	prob = TTX.prob;
	hit_radius = TTX.hit_radius;
	ammo_time = TTX.ammo_time / dt;
	max_rockets = TTX.max_rockets;
	sigma_nav = Vector3D(TTX.sigma_nav,TTX.sigma_nav,TTX.sigma_nav);

	alpha = 0;
	num_rockets = num_channels;
	channels_free = num_channels;
	reload_time = 0;

	for (int i = 0; i < num_channels; i++)
		channels.push_back(Channel());
}

ZRK::~ZRK()
{
	channels.clear();
}

void ZRK::FindFirePoint(Vector3D p_pos, Vector3D p_spd, Vector3D &dot, double &time)
{
	double Vx = p_spd.x, Vy = p_spd.y; //Vz = 0  

	Vector3D tgt = p_pos - pos;

	//   
	double a = (Vx*Vx + Vy*Vy - rocketV*rocketV);
	double b = (tgt.x*Vx + tgt.y*Vy);
	double c = (tgt.x*tgt.x + tgt.y*tgt.y + tgt.z*tgt.z);

	double D = b*b - a*c;

	//   (   /  )
	if (D < 0)
	{
		time = -1;
		return;
	}
	double time1,time2;
	time1 = (-b + sqrt(D))/a;
	time2 = (-b - sqrt(D))/a;

	time = -1;
	//     (  )
	if (time1 > time2)
	{
		if (time1 > 0) time = time1; 
		if (time2 > 0) time = time2;
	}
	else
	{
		if (time2 > 0)	time = time2;
		if (time1 > 0)	time = time1;
	} 

	dot = pos + tgt + p_spd*time;
}

int ZRK::InFireZone(Plane &plane)
{
	// ,       .    
	//      ,      
	Vector3D dot(0,0,0);
	double time;	

	FindFirePoint(plane.pos, plane.speed, dot, time);

	if (time < 0) 
        return 0;
	if (!CheckRange(dot-pos))
		return TryToMoveZone(plane);    //         ,       
	else
		return 1;
}


int ZRK::CheckRange(const Vector3D &point)
{
    // ,        
	if (point.length() < minR || point.length() > maxR)
		return 0;
	double beta = atan2(point.y, point.x);
	return insideAngles(alpha - gamma/2, beta, alpha + gamma/2);
}

ZRK::Ukaz(Plane *plane, Vector3D error)
{
	//if (!InFireZone(*plane)) 
 //       AfxMessageBox("Obman!!!");

	for (Channels::iterator i = channels.begin(); i != channels.end(); i++)
	{
		RocketList::iterator rt;
		if (!i->busy)
		{
			Vector3D total_err = Norm3D(sigma_nav) + error;
			Vector3D tgtPoint = plane->pos + total_err;
			Vector3D dot;
			double time;
			FindFirePoint(tgtPoint, plane->speed, dot, time);
			rockets.push_back(Rocket(plane, ZRK::pos, dot, rocketV));
			rt = rockets.end();
			rt--;
			i->Set(dot, total_err, &(*rt));
			channels_free--;
			num_rockets--;

			CString tmp;
			tmp.Format("       %i.",plane->GetId());
			WriteLog(tmp);

			return 1;
		}
	}
	return 0;
}

int ZRK::TryToMoveZone(Plane &plane)
{
	//      
	double maxNewAlpha = PI, minNewAlpha = -PI;
	int freezone = 1;
	double angle = 0;
	for (Channels::const_iterator i = channels.begin(); i != channels.end(); i++)
	{
		if (i->busy)
		{
			freezone = 0;
			double beta = atan2(i->pos.y-pos.y, i->pos.x-pos.x);
			angle += normalizeAngle(beta);
			if (beta + gamma/2 < maxNewAlpha) maxNewAlpha = beta + gamma/2;
			if (beta - gamma/2 > minNewAlpha) minNewAlpha = beta - gamma/2;
		}
	}

	Vector3D dot(0,0,0);
	double time, beta;		
	FindFirePoint(plane.pos, plane.speed, dot, time);

	if (time > 0)
	{
		beta = atan2(dot.y-pos.y, dot.x-pos.x);
        double tmp;
        tmp = (dot-pos).length();
        if (tmp > minR && tmp < maxR)
            beta = beta;
		if ((freezone || insideAngles(minNewAlpha - gamma/2,beta,maxNewAlpha + gamma/2))
			&& (dot-pos).length() > minR && (dot-pos).length() < maxR)
		{
			alpha = (normalizeAngle(beta) + angle)/(num_channels - channels_free + 1) ;
			return 1;
		}
	}
	return 0;
}


int ZRK::ChannelsFree()
{
	return num_rockets < channels_free ?  num_rockets : channels_free;
}

double ZRK::ChanceToHit(double R)
{
	double chance = prob *(1 - R/hit_radius);
	if (chance < 0) chance = 0;
	return chance;
}

void ZRK::NextTurn()
{	
	/*if(num_channels > 1)
		if(!m_zrk_alive)return;
		else if(!s_zrk_alive)return;
		
	if (!this->IsAlive()) return;
*/

	//  : 
	reload_time++;
	if (reload_time == ammo_time)
	{
		reload_time = 0;
		//
		int need = num_channels - num_rockets;
		if (max_rockets > need)
		{
			max_rockets -= need;
			num_rockets = num_channels;
		}
		else
			num_rockets += max_rockets;
	}

	for (Channels::iterator i = channels.begin(); i != channels.end(); i++)
	{
		if (i->busy)
		{
			if (i->rocket->Move())
			{// 
				if ((ChanceToHit(i->error.length()) > Random()) && 
					(i->rocket->plane->pos - i->pos).length() < (i->rocket->plane->speed - i->rocket->speed).length())
				{
					//:   
					i->rocket->plane->Destroy();
					
					CString tmp;
					tmp.Format("   %i .",i->rocket->plane->GetId());
					WriteLog(tmp);
					
					// 
					i->Free();
					channels_free++;
				}
				else
				{
					//  :   ,   , 
					//   ,   
					if (shift_fire)
					{
						Plane *p = i->rocket->plane;
						i->Free();
						channels_free++;
						if (InFireZone(*p)&& num_rockets > 0) 
						{
							Ukaz(p,Vector3D(0,0,0));

							CString tmp;
							tmp.Format("   %i  .  .",p->GetId());
							WriteLog(tmp);
						}
						else
						{
							Lost(p);
						
							CString tmp;
							tmp.Format("   %i       (  ).",p->GetId());
							WriteLog(tmp);
						}
					}
					else
					{
						Lost(i->rocket->plane);
						i->Free();
						channels_free++;

						CString tmp;
						tmp.Format("   %i  .",i->rocket->plane->GetId());
						WriteLog(tmp);
					}
				}
			}
		}
	}
}

void ZRK::WriteLog(CString str)
{
	CString temp;
	temp.Format("[ %g ]: (%i-) ",TIME,num_channels);
	temp += str;
	Print((LPCTSTR)temp);
}

